from abc import ABC, abstractmethod
from typing import Any, Awaitable, Callable, Dict, Generic, TypeVar

from humps import decamelize

from kilroy_server_py_utils.categorizable import Categorizable
from kilroy_server_py_utils.utils import classproperty, noop, normalize

StateType = TypeVar("StateType")
ParameterType = TypeVar("ParameterType")


class ParameterGetError(Exception):
    pass


class ParameterSetError(Exception):
    pass


class Parameter(Categorizable, Generic[StateType, ParameterType], ABC):
    async def get(self, state: StateType) -> ParameterType:
        try:
            return await self._get(state)
        except Exception as e:
            raise ParameterGetError() from e

    async def set(
        self,
        state: StateType,
        value: ParameterType,
    ) -> Callable[[], Awaitable]:
        if (await self.get(state)) == value:
            return noop
        try:
            return await self._set(state, value)
        except Exception as e:
            raise ParameterSetError() from e

    async def _get(self, state: StateType) -> ParameterType:
        return getattr(state, decamelize(self.name))

    async def _set(
        self,
        state: StateType,
        value: ParameterType,
    ) -> Callable[[], Awaitable]:
        name = decamelize(self.name)
        original_value = getattr(state, name)

        async def undo():
            setattr(state, name, original_value)

        setattr(state, name, value)
        return undo

    @classproperty
    def category(cls) -> str:
        return cls.name

    @classproperty
    def name(cls) -> str:
        class_name: str = cls.__name__
        name = class_name.removesuffix("Parameter").removeprefix("Parameter")
        return normalize(name) or "parameter"

    @classproperty
    @abstractmethod
    def schema(cls) -> Dict[str, Any]:
        pass



from abc import ABC, abstractmethod
from asyncio import Event, Lock, Queue
from contextlib import asynccontextmanager
from typing import AsyncIterable, Dict, Generic, Type, TypeVar
from uuid import UUID, uuid4

T = TypeVar("T")
ObservableType = TypeVar("ObservableType", bound="Observable")
ObservableImplType = TypeVar("ObservableImplType", bound="ObservableImpl")


class NotInitializedError(Exception):
    pass


class FetchableObservable(Generic[T], ABC):
    @abstractmethod
    async def fetch(self) -> T:
        pass

    @abstractmethod
    def subscribe(self) -> AsyncIterable[T]:
        pass


class ReadableObservable(FetchableObservable[T], Generic[T], ABC):
    @abstractmethod
    async def get(self) -> T:
        pass

    @property
    def fetch_only(self) -> FetchableObservable[T]:
        return FetchOnlyObservableWrapper(self)


class Observable(ReadableObservable[T], Generic[T], ABC):
    @abstractmethod
    async def set(self, value: T) -> None:
        pass

    @classmethod
    async def build(
        cls: Type[ObservableType], *args, **kwargs
    ) -> ObservableType:
        return await ObservableImpl.build(*args, **kwargs)

    @property
    def read_only(self) -> ReadableObservable[T]:
        return ReadOnlyObservableWrapper(self)


class ObservableImpl(Observable[T], Generic[T]):
    _value_lock: Lock
    _queue_lock: Lock
    _ready_event: Event
    _value: T
    _queues: Dict[UUID, Queue[T]]

    def __init__(
        self,
        *args,
        value_lock: Lock,
        queue_lock: Lock,
        ready_event: Event,
        **kwargs,
    ) -> None:
        self._value_lock = value_lock
        self._queue_lock = queue_lock
        self._ready_event = ready_event
        if len(args) > 0:
            self._value = args[0]
            self._ready_event.set()
        elif "value" in kwargs:
            self._value = kwargs["value"]
            self._ready_event.set()
        self._queues = {}

    @classmethod
    async def build(
        cls: Type[ObservableImplType], *args, **kwargs
    ) -> ObservableImplType:
        return cls(
            *args,
            value_lock=Lock(),
            queue_lock=Lock(),
            ready_event=Event(),
            **kwargs,
        )

    @asynccontextmanager
    async def _create_queue(self) -> Queue[T]:
        queue_id = uuid4()
        queue = Queue()

        async with self._queue_lock:
            self._queues[queue_id] = queue

        yield queue

        async with self._queue_lock:
            self._queues.pop(queue_id)

    async def _notify(self) -> None:
        async with self._queue_lock:
            for queue in self._queues.values():
                await queue.put(self._value)

    async def set(self, value: T) -> None:
        async with self._value_lock:
            if not self._ready_event.is_set():
                self._value = value
                await self._notify()
                self._ready_event.set()
            elif self._value != value:
                self._value = value
                await self._notify()

    async def get(self) -> T:
        async with self._value_lock:
            if not self._ready_event.is_set():
                raise NotInitializedError()
            return self._value

    async def fetch(self) -> T:
        await self._ready_event.wait()
        return await self.get()

    async def subscribe(self) -> AsyncIterable[T]:
        async with self._create_queue() as queue:
            while (message := await queue.get()) is not None:
                yield message


class FetchOnlyObservableWrapper(FetchableObservable[T], Generic[T]):
    _observable: FetchableObservable[T]

    def __init__(self, observable: FetchableObservable[T]) -> None:
        self._observable = observable

    async def fetch(self) -> T:
        return await self._observable.fetch()

    def subscribe(self) -> AsyncIterable[T]:
        return self._observable.subscribe()


class ReadOnlyObservableWrapper(ReadableObservable[T], Generic[T]):
    _observable: ReadableObservable[T]

    def __init__(self, observable: ReadableObservable[T]) -> None:
        self._observable = observable

    async def get(self) -> T:
        return await self._observable.get()

    async def fetch(self) -> T:
        return await self._observable.fetch()

    def subscribe(self) -> AsyncIterable[T]:
        return self._observable.subscribe()



from abc import ABC
from typing import Any, Awaitable, Callable, Dict, Generic, Type, TypeVar

from humps import decamelize

from kilroy_server_py_utils.categorizable import Categorizable
from kilroy_server_py_utils.configurable import Configurable
from kilroy_server_py_utils.parameters.base import Parameter
from kilroy_server_py_utils.utils import classproperty, get_generic_args, noop

StateType = TypeVar("StateType")
ParameterType = TypeVar("ParameterType")
ConfigurableType = TypeVar("ConfigurableType", bound=Configurable)
CategorizableType = TypeVar("CategorizableType", bound=Categorizable)


class NestedParameter(
    Parameter[StateType, Dict[str, Any]],
    Generic[StateType, ConfigurableType],
    ABC,
):
    async def _get(self, state: StateType) -> Dict[str, Any]:
        configurable = await self._get_configurable(state)
        return await configurable.config.json.fetch()

    async def _set(
        self, state: StateType, value: Dict[str, Any]
    ) -> Callable[[], Awaitable]:
        configurable = await self._get_configurable(state)
        original_value = await configurable.config.json.get()

        async def undo():
            await configurable.config.set(original_value)

        await configurable.config.set(value)
        return undo

    async def _get_configurable(self, state: StateType) -> ConfigurableType:
        return getattr(state, decamelize(self.name))

    @classproperty
    def configurable_class(cls) -> Type[ConfigurableType]:
        # noinspection PyUnresolvedReferences
        return get_generic_args(cls, NestedParameter)[1]

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": cls.configurable_class.properties_schema,
        }


class CategorizableBasedParameter(
    Parameter[StateType, Dict[str, Any]],
    Generic[StateType, CategorizableType],
    ABC,
):
    async def _get(self, state: StateType) -> Dict[str, Any]:
        categorizable = await self._get_categorizable(state)
        category = categorizable.category
        if isinstance(categorizable, Configurable):
            return {
                "type": category,
                "config": await categorizable.config.json.get(),
            }
        return {"type": category}

    async def _set(
        self, state: StateType, value: Dict[str, Any]
    ) -> Callable[[], Awaitable]:
        current = await self._get_categorizable(state)
        category = value.get("type", current.category)

        current = await self._get_categorizable(state)
        if current.category == category:
            if isinstance(current, Configurable):
                original_value = await current.config.json.get()

                async def undo():
                    await current.config.set(original_value)

                await current.config.set(value.get("config", {}))
                return undo

            return noop

        if isinstance(current, Configurable):
            await current.cleanup()

        params = await self._get_params(state, category)
        cls = self.categorizable_base_class.for_category(category)
        if issubclass(cls, Configurable):
            instance = await cls.build(**params)
            await instance.init()
            original_config = await instance.config.json.get()

            async def undo():
                await instance.config.set(original_config)
                await self._set_categorizable(state, current)

            await instance.config.set(value.get("config", {}))
            # noinspection PyTypeChecker
            await self._set_categorizable(state, instance)

            return undo
        else:
            instance = cls(**params)

            async def undo():
                await self._set_categorizable(state, current)

            await self._set_categorizable(state, instance)

            return undo

    async def _get_categorizable(self, state: StateType) -> CategorizableType:
        return getattr(state, decamelize(self.name))

    async def _set_categorizable(
        self, state: StateType, value: CategorizableType
    ) -> None:
        setattr(state, decamelize(self.name), value)

    async def _get_params(
        self, state: StateType, category: str
    ) -> Dict[str, Any]:
        all_params = getattr(state, f"{decamelize(self.name)}s_params")
        return all_params.get(category, {})

    @classproperty
    def categorizable_base_class(cls) -> Type[CategorizableType]:
        # noinspection PyUnresolvedReferences
        return get_generic_args(cls, CategorizableBasedParameter)[1]

    @classproperty
    def schema(cls) -> Dict[str, Any]:
        possible_properties = []
        for category in cls.categorizable_base_class.all_categories:
            properties = {
                "type": {"type": "string", "const": category},
            }
            sampler_class = cls.categorizable_base_class.for_category(category)
            if issubclass(sampler_class, Configurable):
                properties["config"] = {
                    "type": "object",
                    "properties": sampler_class.properties_schema,
                }
            possible_properties.append(properties)
        return {
            "type": "object",
            "oneOf": [
                {"type": "object", "properties": properties}
                for properties in possible_properties
            ],
        }



from typing import Any, Dict

from jsonschema.exceptions import SchemaError
from jsonschema.validators import Draft202012Validator
from pydantic import BaseModel, Extra, root_validator


class JSONSchema(BaseModel, extra=Extra.allow):
    __root__: Dict[str, Any]

    @root_validator(pre=True)
    def validate(cls, values) -> Dict[str, Any]:
        try:
            Draft202012Validator.check_schema(values)
        except SchemaError as e:
            raise ValueError(
                "Schema is not a valid JSON Schema 2020-12."
            ) from e
        if "type" not in values:
            raise ValueError("Schema should have a type field.")
        elif values["type"] != "object":
            raise ValueError("Only object types are allowed.")
        return {"__root__": values}
